structure Ast =
struct

  exception Exit_Loop

  type sslist = (string * string) list
  
  type reservedchar = Hat | Underbar | Dq
  type reservedstr = Vdash | Mapsto | Rbigarrow 
                   | Lbigarrow | Rarrow | Larrow
                   | In | Ni 
  type angle = Langle | Rangle

  type reserved = RChar of reservedchar
                | RStr of reservedstr 
                | Ang of angle
                | Escape 
                

  type escaped = Escaped of string * reservedchar 

  type latex = LatexEsc of escaped 
             | LatexEscChar of string

  type primary = Var of sslist 
               | PrRstr of reservedstr 
               | PrAng of angle
               | Mathsym of string
  
  type updown = SupPr of primary
              | SupLa of latex
              | SubPr of primary
              | SubLa of latex

  type exp = Prim of primary 
           | Updn of updown 

  type content = RuleCont_Exp of exp
               | RuleCont_Latex of latex

  type rule = RuleAxiom of content list
            | RuleBasic of content list * content list 
            | RuleNest of rule * content list 
  
  type syntex = Exp of exp 
              | Rule of rule 
  
 
  type toplevel = Syntex of syntex * (int*int) 
                | Latex of latex * (int*int)

  type top = Top of toplevel list

val nt = ref ","
val tl = ref "."
val cm = ref "/"

fun print_reservedstr oc rs =
  case rs of 
    Vdash => output_string oc "\\vdash"
  | Mapsto => output_string oc "\\mapsto"
  | Rbigarrow => output_string oc "\\Rightarrow"
  | Lbigarrow => output_string oc "\\Leftarrow"
  | Rarrow => output_string oc "\\rightarrow"
  | Larrow => output_string oc "\\leftarrow"
  | In => output_string oc "\\in"
  | Ni => output_string oc "\\ni"

fun print_angle_res oc an =
  case an of
    Langle => output_string oc "\\langle"
  | Rangle => output_string oc "\\rangle"

fun print_angle oc an =
  case an of
    Langle => output_string oc "<"
  | Rangle => output_string oc ">"

fun print_normal oc s =
  if (s="") then () else (
  let val ch = String.sub s 0 1
      val len = String.length s
  in (case ch of
       "#" => output_string oc "#"
     | "&" => output_string oc "&"
     | "$" => output_string oc "$"
     | "{" => output_string oc "{"
     | "}" => output_string oc "}"
     | "%" => output_string oc "%"
     | "~" => output_string oc "~"
     | _ => output_string oc ch); print_normal oc (String.sub s 1 (len-1))
  end)

fun print_withfont oc s =
  if (s="") then () else (
  let val ch = String.sub s 0 1
      val len = String.length s
  in (case ch of
       "#" => output_string oc "\\#"
     | "&" => output_string oc "\\&"
     | "$" => output_string oc "\\$"
     | "{" => output_string oc "\\{"
     | "}" => output_string oc "\\}"
     | "%" => output_string oc "\\%"
     | "~" => output_string oc "\\char`\\~"
     | _ => output_string oc ch); print_withfont oc (String.sub s 1 (len-1))
  end)
 
fun print_escape oc (Escaped(s,rc)) =
  case rc of
    Hat => (case s of
              "nt" => output_string oc "\mbox{\\it \\char`\\^}"
            | "tl" => output_string oc "\mbox{\\tt \\char`\\^}"
            | "rm" => output_string oc "\\mbox{\\char`\\^}"
            | _ => output_string oc "\\char`\\^")
  | Underbar => (case s of
                   "nt" => output_string oc "\mbox{\\it \\char`\\_}"
                 | "tl" => output_string oc "\mbox{\\tt \\char`\\_}"
                 | "rm" => output_string oc "\\mbox{\\char`\\_}"
                 | _ => output_string oc "\\char`\\_")
  | Dq => (case s of
             "nt" => output_string oc "{\\it \"}"
           | "tl" => output_string oc "{\\tt \"}"
           | "rm" => output_string oc "\\mbox{\"}"
           | _ => output_string oc "\"")

fun print_latex oc s =
  case s of
    LatexEsc(a) => print_escape oc a
  | LatexEscChar(s) => 
     (case s of 
        "nt" => output_string oc "\\mbox{\\it \\char`\\\\}"
      | "tl" => output_string oc "\\mbox{\\tt \\char`\\\\}"
      | "rm" => output_string oc "\\mbox{\\char`\\\\}"
      | _ => output_string oc "\\")

fun print_mathsym oc ms = 
  output_string oc ms

fun print_var oc slist =
  case slist of 
    [] => ()
  | (cont,ty)::tl =>(case ty of
                       "normal" => print_normal oc cont; print_var oc tl
                     | "nt" => output_string oc "\\mbox{\\it ";
                               print_withfont oc cont;  
                               output_string oc "}";
                               print_var oc tl
                     | "tl" => output_string oc "\\mbox{\\tt ";
                               print_withfont oc cont;  
                               output_string oc "}";
                               print_var oc tl
                     | "rm" => output_string oc "\\mbox{";
                               print_withfont oc cont;  
                               output_string oc "}";
                               print_var oc tl
                     | _ => print_var oc tl)

fun print_primary oc p =
  case p of
    Var(s) => print_var oc s
  | PrRstr(rs) => print_reservedstr oc rs  
  | PrAng(an) => print_angle_res oc an
  | Mathsym(ms) => print_mathsym oc ms

fun print_updown oc ud =
  case ud of
    SupPr(pr) => output_string oc "^";
                 print_primary oc pr
  | SupLa(la) => output_string oc "^";
                 print_latex oc la
  | SubPr(pr) => output_string oc "_";
                 print_primary oc pr
  | SubLa(la) => output_string oc "_";
                 print_latex oc la

fun print_exp oc e =
  case e of
    Prim(p) => print_primary oc p
  | Updn(ud) => print_updown oc ud 

fun whats_l (LatexEscChar(s)) = "\\"
  | whats_l (_) = "@"

fun print_clist oc cl =
  case cl of
    [] => ()
  | (RuleCont_Exp(e))::tl => print_exp oc e;
                                     output_string oc " ";
                                     print_clist oc tl
  | (RuleCont_Latex(l))::tl => print_latex oc l; 
                                       if ((whats_l l)="\\") then print_clist oc tl
                                         else (output_string oc " ";
                                               print_clist oc tl)

fun print_rule oc r =
  case r of
    RuleAxiom(cl) => 
      output_string oc "\\begin{array}{c}\n";
      output_string oc "\\infer{";
      print_clist oc cl;
      output_string oc "}{}\n";
      output_string oc "\\end{array}"
  | RuleBasic(cl1,cl2) =>
      output_string oc "\\begin{array}{c}\n";
      output_string oc "\\infer{";
      print_clist oc cl2;
      output_string oc "}{";
      print_clist oc cl1;
      output_string oc "}\n";
      output_string oc "\\end{array}"
  | RuleNest(rule,cl) =>
      output_string oc "\\begin{array}{c}\n";
      output_string oc "\\infer{";
      print_clist oc cl;
      output_string oc "}{";
      print_rule oc rule;
      output_string oc "}\n";
      output_string oc "\\end{array}"

fun print_syntex oc syn = 
 case syn of 
    Exp(e) => print_exp oc e
  | Rule(r) => print_rule oc r

fun print_top oc t buf l = 
  case t of
    [] => output_string oc (String.sub buf l ((String.length buf)-(l+1)))  
  | (Syntex(s,(x1,x2)))::tl => (if (l=x1) then print_syntex oc s
                                          else (output_string oc (String.sub buf l (x1-l));
                                                print_syntex oc s));
                              print_top oc tl buf x2 
  | (Latex(s,(x1,x2)))::tl => (if (l=x1) then print_latex oc s
                                         else (output_string oc (String.sub buf l (x1-l));
                                               print_latex oc s));
                               print_top oc tl buf x2 

fun runtop oc (Top(t)) fid =
  let val ic = open_in fid
      val buff = ref ""
  in 
    (while true do 
      (let val str = input_line ic
       in buff:=(!buff)^(str^"\n")
       end) handle End_of_file => raise Exit_Loop
     end) handle Exit_Loop =>
     close_in ic;
     print_top oc t (!buff) 0
  end

end
